/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package apiclient

import (
	"io/ioutil"
	"os"
	"path/filepath"
	"strings"
	"testing"
	"time"

	"github.com/sirupsen/logrus"
	"github.com/sirupsen/logrus/hooks/test"

	api_v1 "k8s.io/api/core/v1"
	meta_v1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

func TestVCLChanged(t *testing.T) {
	logger, hook := test.NewNullLogger()
	updater := &SpecUpdater{
		log: logger,
	}
	updater.log.Level = logrus.TraceLevel
	cfgMap := &api_v1.ConfigMap{
		ObjectMeta: meta_v1.ObjectMeta{
			Namespace: "default",
			Name:      "vcl-cfgmap",
		},
	}
	changed, err := updater.VCLChanged(cfgMap)
	if err != nil {
		t.Fatalf("VCLChanged(%v): %v", cfgMap, err)
	}
	if changed {
		t.Errorf("VCLChanged(%v) got: true, want: false", cfgMap)
	}
	if hook.LastEntry().Level != logrus.WarnLevel {
		t.Errorf("VCLChanged(%v) log level got: %v, want: WarnLevel",
			cfgMap, hook.LastEntry().Level)
	}
	msg := "ConfigMap default/vcl-cfgmap has no data"
	if hook.LastEntry().Message != msg {
		t.Errorf("VCLChanged(%v) log message got: %s, want: %s",
			cfgMap, hook.LastEntry().Message, msg)
	}

	vclMap, err := updater.updatedVCLMap(cfgMap)
	if err != nil {
		t.Fatalf("updatedVCLMap(%v): %v", cfgMap, err)
	}
	if vclMap != nil {
		t.Fatalf("updatedVCLMap(%v): got %v, want: nil", cfgMap, vclMap)
	}

	tmpdir, err := ioutil.TempDir("", "k8s-vcl-reloader")
	if err != nil {
		t.Fatalf("Cannot create tmp directory: %v", err)
	}
	defer os.RemoveAll(tmpdir)
	tmpdir = filepath.Join(tmpdir, "apiclient_spec_test")
	if err = os.Mkdir(tmpdir, 0755); err != nil {
		t.Fatalf("Cannot make tmp directory %s: %s", tmpdir, err)
	}
	updater.cfg.VCLRoot = tmpdir

	cfgMap.Data = make(map[string]string)
	cfgMap.BinaryData = make(map[string][]byte)
	cfgMap.Data["foo"] = ""
	cfgMap.Data["bar"] = ""
	cfgMap.BinaryData["baz"] = []byte("")
	cfgMap.BinaryData["quux"] = []byte("")
	changed, err = updater.VCLChanged(cfgMap)
	if err != nil {
		t.Fatalf("VCLChanged(%v): %v", cfgMap, err)
	}
	if !changed {
		t.Errorf("VCLChanged(%v) got: false, want: true", cfgMap)
	}

	for _, file := range []string{"foo", "bar", "baz", "quux"} {
		path := filepath.Join(tmpdir, file)
		if err = ioutil.WriteFile(path, []byte{}, 0644); err != nil {
			t.Fatalf("Cannot write %s: %v", path, err)
		}
		defer os.Remove(path)
	}
	changed, err = updater.VCLChanged(cfgMap)
	if err != nil {
		t.Fatalf("VCLChanged(%v): %v", cfgMap, err)
	}
	if !changed {
		t.Errorf("VCLChanged(%v) got: false, want: true", cfgMap)
	}
	vclMap, err = updater.updatedVCLMap(cfgMap)
	if err != nil {
		t.Fatalf("updatedVCLMap(%v): %v", cfgMap, err)
	}
	if vclMap == nil {
		t.Errorf("updatedVCLMap(%v): got nil, want: non-nil", cfgMap)
	}
	if len(vclMap) != 4 {
		t.Errorf("updatedVCLMap(%v): got len(map) = %d, want: 4",
			cfgMap, len(vclMap))
	}
	updater.vcls = vclMap
	changed, err = updater.VCLChanged(cfgMap)
	if err != nil {
		t.Fatalf("VCLChanged(%v): %v", cfgMap, err)
	}
	if changed {
		t.Errorf("VCLChanged(%v) got: true, want: false", cfgMap)
	}

	path := filepath.Join(tmpdir, "foo")
	if err = os.Chtimes(path, time.Now(), time.Now()); err != nil {
		t.Fatalf("os.Chtimes(%s): %v", path, err)
	}
	changed, err = updater.VCLChanged(cfgMap)
	if err != nil {
		t.Fatalf("VCLChanged(%v): %v", cfgMap, err)
	}
	if !changed {
		t.Errorf("VCLChanged(%v) got: false, want: true", cfgMap)
	}
	if hook.LastEntry().Level != logrus.TraceLevel {
		t.Errorf("VCLChanged(%v) log level got: %v, want: TraceLevel",
			cfgMap, hook.LastEntry().Level)
	}
	if !strings.Contains(hook.LastEntry().Message, "foo new mtime") {
		t.Errorf("VCLChanged(%v) log message got: %s, should contain: "+
			"foo new mtime", cfgMap, hook.LastEntry().Message)
	}
	vclMap, err = updater.updatedVCLMap(cfgMap)
	if err != nil {
		t.Fatalf("updatedVCLMap(%v): %v", cfgMap, err)
	}
	if vclMap == nil {
		t.Errorf("updatedVCLMap(%v): got nil, want: non-nil", cfgMap)
	}
	if len(vclMap) != 4 {
		t.Errorf("updatedVCLMap(%v): got len(map) = %d, want: 4",
			cfgMap, len(vclMap))
	}
	updater.vcls = vclMap
	changed, err = updater.VCLChanged(cfgMap)
	if err != nil {
		t.Fatalf("VCLChanged(%v): %v", cfgMap, err)
	}
	if changed {
		t.Errorf("VCLChanged(%v) got: true, want: false", cfgMap)
	}
	if hook.LastEntry().Level != logrus.TraceLevel {
		t.Errorf("VCLChanged(%v) log level got: %v, want: TraceLevel",
			cfgMap, hook.LastEntry().Level)
	}
	if !strings.Contains(hook.LastEntry().Message, "foo new mtime") {
		t.Errorf("VCLChanged(%v) log message got: %s, should contain: "+
			"foo new mtime", cfgMap, hook.LastEntry().Message)
	}

	cfgMap.Data["4711"] = ""
	changed, err = updater.VCLChanged(cfgMap)
	if err != nil {
		t.Fatalf("VCLChanged(%v): %v", cfgMap, err)
	}
	if changed {
		t.Errorf("VCLChanged(%v) got: true, want: false", cfgMap)
	}
	if hook.LastEntry().Level != logrus.TraceLevel {
		t.Errorf("VCLChanged(%v) log level got: %v, want: TraceLevel",
			cfgMap, hook.LastEntry().Level)
	}
	if !strings.HasPrefix(hook.LastEntry().Message, "File not added: ") {
		t.Errorf("VCLChanged(%v) log message got: %s, should contain: "+
			"File not added: ", cfgMap, hook.LastEntry().Message)
	}

	path = filepath.Join(tmpdir, "4711")
	if err = ioutil.WriteFile(path, []byte{}, 0644); err != nil {
		t.Fatalf("Cannot write %s: %v", path, err)
	}
	defer os.Remove(path)
	changed, err = updater.VCLChanged(cfgMap)
	if err != nil {
		t.Fatalf("VCLChanged(%v): %v", cfgMap, err)
	}
	if !changed {
		t.Errorf("VCLChanged(%v) got: false, want: true", cfgMap)
	}
	if hook.LastEntry().Level != logrus.TraceLevel {
		t.Errorf("VCLChanged(%v) log level got: %v, want: TraceLevel",
			cfgMap, hook.LastEntry().Level)
	}
	if !strings.HasPrefix(hook.LastEntry().Message, "File added: ") {
		t.Errorf("VCLChanged(%v) log message got: %s, should contain: "+
			"File added: ", cfgMap, hook.LastEntry().Message)
	}
	vclMap, err = updater.updatedVCLMap(cfgMap)
	if err != nil {
		t.Fatalf("updatedVCLMap(%v): %v", cfgMap, err)
	}
	if vclMap == nil {
		t.Errorf("updatedVCLMap(%v): got nil, want: non-nil", cfgMap)
	}
	if len(vclMap) != 5 {
		t.Errorf("updatedVCLMap(%v): got len(map) = %d, want: 5",
			cfgMap, len(vclMap))
	}
	updater.vcls = vclMap
	changed, err = updater.VCLChanged(cfgMap)
	if err != nil {
		t.Fatalf("VCLChanged(%v): %v", cfgMap, err)
	}
	if changed {
		t.Errorf("VCLChanged(%v) got: true, want: false", cfgMap)
	}

	delete(cfgMap.Data, "4711")
	changed, err = updater.VCLChanged(cfgMap)
	if err != nil {
		t.Fatalf("VCLChanged(%v): %v", cfgMap, err)
	}
	if changed {
		t.Errorf("VCLChanged(%v) got: true, want: false", cfgMap)
	}
	if hook.LastEntry().Level != logrus.TraceLevel {
		t.Errorf("VCLChanged(%v) log level got: %v, want: TraceLevel",
			cfgMap, hook.LastEntry().Level)
	}
	msg = "File not deleted: 4711"
	if hook.LastEntry().Message != msg {
		t.Errorf("VCLChanged(%v) log message got: %s, want: %s",
			cfgMap, hook.LastEntry().Message, msg)
	}

	if err = os.Remove(path); err != nil {
		t.Fatalf("os.Remove(%s): %v", path, err)
	}
	changed, err = updater.VCLChanged(cfgMap)
	if err != nil {
		t.Fatalf("VCLChanged(%v): %v", cfgMap, err)
	}
	if !changed {
		t.Errorf("VCLChanged(%v) got: false, want: true", cfgMap)
	}
	if hook.LastEntry().Level != logrus.TraceLevel {
		t.Errorf("VCLChanged(%v) log level got: %v, want: TraceLevel",
			cfgMap, hook.LastEntry().Level)
	}
	if hook.LastEntry().Message != "File deleted: 4711" {
		t.Errorf("VCLChanged(%v) log message got: %s, want: "+
			"File deleted: 4711", cfgMap, hook.LastEntry().Message)
	}
	vclMap, err = updater.updatedVCLMap(cfgMap)
	if err != nil {
		t.Fatalf("updatedVCLMap(%v): %v", cfgMap, err)
	}
	if vclMap == nil {
		t.Errorf("updatedVCLMap(%v): got nil, want: non-nil", cfgMap)
	}
	if len(vclMap) != 4 {
		t.Errorf("updatedVCLMap(%v): got len(map) = %d, want: 4",
			cfgMap, len(vclMap))
	}
	updater.vcls = vclMap
	changed, err = updater.VCLChanged(cfgMap)
	if err != nil {
		t.Fatalf("VCLChanged(%v): %v", cfgMap, err)
	}
	if changed {
		t.Errorf("VCLChanged(%v) got: true, want: false", cfgMap)
	}
}

func flushFile(t *testing.T, path string) {
	if file, err := os.Open(path); err != nil {
		t.Fatalf("Open(%s): %v", path, err)
	} else {
		if err = file.Sync(); err != nil {
			t.Fatalf("Sync(%s): %v", path, err)
		}
		if err = file.Close(); err != nil {
			t.Fatalf("Close(%s): %v", path, err)
		}
	}
}

func TestGenVCLChanged(t *testing.T) {
	tmpdir, err := ioutil.TempDir("", "k8s-vcl-reloader")
	if err != nil {
		t.Fatalf("Cannot create tmp directory: %v", err)
	}
	defer os.RemoveAll(tmpdir)

	logger, hook := test.NewNullLogger()
	updater := &SpecUpdater{
		log:     logger,
		genvcls: make(map[string]time.Time, 4),
	}
	updater.log.Level = logrus.TraceLevel

	for _, file := range []string{"foo", "bar", "baz", "quux"} {
		updater.genvcls[filepath.Join(tmpdir, file+".vcl")] =
			time.Time{}
	}
	changed, err := updater.genVCLChanged()
	if err != nil {
		t.Fatal("genVCLChanged() (no files): ", err)
	}
	if changed {
		t.Error("genVCLChanged(): got: true, want: false")
	}
	if hook.LastEntry().Level != logrus.WarnLevel {
		t.Errorf("genVCLChanged() log level got: %v, want: WarnLevel",
			hook.LastEntry().Level)
	}
	pfx := "Generated VCL not found:"
	if !strings.HasPrefix(hook.LastEntry().Message, pfx) {
		t.Errorf("genVCLChanged() log message got: \"%s\", "+
			"want prefix: %s", hook.LastEntry().Message, pfx)
	}

	for path := range updater.genvcls {
		if err = ioutil.WriteFile(path, []byte{}, 0644); err != nil {
			t.Fatalf("Cannot write %s: %v", path, err)
		}
		flushFile(t, path)
		defer os.Remove(path)
	}
	changed, err = updater.genVCLChanged()
	if err != nil {
		t.Fatal("genVCLChanged(): ", err)
	}
	if !changed {
		t.Error("genVCLChanged() got: false, want: true")
	}

	if updater.genvcls, err = updater.updatedGenVCLMap(); err != nil {
		t.Fatal("updatedGenVCLMap() :", err)
	}
	changed, err = updater.genVCLChanged()
	if err != nil {
		t.Fatal("genVCLChanged(): ", err)
	}
	if changed {
		t.Error("genVCLChanged() got: true, want: false")
	}

	paths := make([]string, len(updater.genvcls))
	i := 0
	for path := range updater.genvcls {
		paths[i] = path
		i++
	}
	for _, path := range paths {
		if err = ioutil.WriteFile(path, []byte{47}, 0644); err != nil {
			t.Fatalf("Cannot write %s: %v", path, err)
		}
		flushFile(t, path)
		changed, err = updater.genVCLChanged()
		if err != nil {
			t.Fatal("genVCLChanged(): ", err)
		}
		if !changed {
			t.Error("genVCLChanged() got: false, want: true")
		}

		if updater.genvcls, err = updater.updatedGenVCLMap(); err != nil {
			t.Fatal("updatedGenVCLMap() :", err)
		}
		changed, err = updater.genVCLChanged()
		if err != nil {
			t.Fatal("genVCLChanged(): ", err)
		}
		if changed {
			t.Error("genVCLChanged() got: true, want: false")
		}
	}
}
