/*
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package varnish

import (
	"sync"

	"github.com/prometheus/client_golang/prometheus"
)

const (
	namespace = "vclreloader"
	subsystem = "varnish"
)

type instanceMetrics struct {
	updates         prometheus.Counter
	updateErrs      prometheus.Counter
	tmplFails       prometheus.Counter
	vsmFails        prometheus.Counter
	connectFails    prometheus.Counter
	vclLoads        prometheus.Counter
	vclLoadErrs     prometheus.Counter
	templateGens    prometheus.Counter
	templateErrs    prometheus.Counter
	connectLatency  prometheus.Summary
	vclLoadLatency  prometheus.Summary
	pings           prometheus.Counter
	pingFails       prometheus.Counter
	panics          prometheus.Counter
	restarts        prometheus.Counter
	restartFails    prometheus.Counter
	childRunning    prometheus.Counter
	childNotRunning prometheus.Counter
	vclDiscards     prometheus.Counter
	monitorChecks   prometheus.Counter
}

var (
	beEndpsGauge = prometheus.NewGauge(prometheus.GaugeOpts{
		Namespace: namespace,
		Subsystem: subsystem,
		Name:      "backend_endpoints",
		Help: "Current number of Service endpoints configured " +
			"as Varnish backends",
	})

	monResultCtr = prometheus.NewCounterVec(prometheus.CounterOpts{
		Namespace: namespace,
		Subsystem: subsystem,
		Name:      "monitor_result_total",
		Help:      "Total number of monitor results",
	}, []string{"status", "result"})

	metrics = &instanceMetrics{
		updates: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "updates_total",
			Help:      "Total number of attempted updates",
		}),
		updateErrs: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "update_errors_total",
			Help:      "Total number of update errors",
		}),
		tmplFails: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "template_errors_total",
			Help:      "Total number of template parse failures",
		}),
		vsmFails: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "vsm_errors_total",
			Help:      "Total number of shared memory attach failures",
		}),
		connectFails: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "admin_connect_fails_total",
			Help:      "Total number of admin connection failures",
		}),
		vclLoads: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "vcl_loads_total",
			Help:      "Total number of successful VCL loads",
		}),
		vclLoadErrs: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "vcl_load_errors_total",
			Help:      "Total number of VCL load errors",
		}),
		templateGens: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "vcl_gens_total",
			Help:      "Total number of successful VCL generations",
		}),
		templateErrs: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "vcl_gen_errors_total",
			Help:      "Total number of VCL generation errors",
		}),
		connectLatency: prometheus.NewSummary(prometheus.SummaryOpts{
			Namespace:  namespace,
			Subsystem:  subsystem,
			Name:       "admin_connect_latency_seconds",
			Help:       "Admin connection latency",
			Objectives: latencyObjectives,
		}),
		vclLoadLatency: prometheus.NewSummary(prometheus.SummaryOpts{
			Namespace:  namespace,
			Subsystem:  subsystem,
			Name:       "vcl_load_latency_seconds",
			Help:       "VCL load latency",
			Objectives: latencyObjectives,
		}),
		pings: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "pings_total",
			Help:      "Total number of successful pings",
		}),
		pingFails: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "ping_errors_total",
			Help:      "Total number of ping errors",
		}),
		panics: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "panics_total",
			Help:      "Total number of panics detected",
		}),
		restarts: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "restarts_total",
			Help:      "Total number of child process restarts",
		}),
		restartFails: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "restart_fails_total",
			Help:      "Total number of child process restart failures",
		}),
		childRunning: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "child_running_total",
			Help: "Total number of monitor runs with the " +
				"child process in the running state",
		}),
		childNotRunning: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "child_not_running_total",
			Help: "Total number of monitor runs with the " +
				"child process not in the running state",
		}),
		vclDiscards: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "vcl_discards_total",
			Help:      "Total number of VCL discards",
		}),
		monitorChecks: prometheus.NewCounter(prometheus.CounterOpts{
			Namespace: namespace,
			Subsystem: subsystem,
			Name:      "monitor_checks_total",
			Help:      "Total number of monitor checks",
		}),
	}

	instMetricsMtx = &sync.Mutex{}

	latencyObjectives = map[float64]float64{
		0.5:   0.001,
		0.9:   0.001,
		0.95:  0.001,
		0.99:  0.001,
		0.999: 0.001,
	}
)

func initMetrics() {
	prometheus.Register(monResultCtr)
	prometheus.Register(beEndpsGauge)
	prometheus.Register(metrics.updates)
	prometheus.Register(metrics.updateErrs)
	prometheus.Register(metrics.vsmFails)
	prometheus.Register(metrics.tmplFails)
	prometheus.Register(metrics.connectFails)
	prometheus.Register(metrics.vclLoads)
	prometheus.Register(metrics.vclLoadErrs)
	prometheus.Register(metrics.templateErrs)
	prometheus.Register(metrics.templateGens)
	prometheus.Register(metrics.connectLatency)
	prometheus.Register(metrics.vclLoadLatency)
	prometheus.Register(metrics.pings)
	prometheus.Register(metrics.pingFails)
	prometheus.Register(metrics.panics)
	prometheus.Register(metrics.restarts)
	prometheus.Register(metrics.restartFails)
	prometheus.Register(metrics.childRunning)
	prometheus.Register(metrics.childNotRunning)
	prometheus.Register(metrics.vclDiscards)
	prometheus.Register(metrics.monitorChecks)
}
